from rich.style import Style
from rich.table import Table
from textual import log
from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.messages import Message
from textual.widget import Widget
from textual.widgets import Button, Footer, Header, Label, Placeholder, Static, Tree

import girok.api.category as category_api
import girok.calendar_cli.utils as calendar_utils
from girok.calendar_cli.calendar_container import CalendarContainer
from girok.calendar_cli.sidebar import CategoryTree, SidebarContainer
from girok.calendar_cli.entity import Category
from girok.constants import Emoji

class CalendarApp(Horizontal):
    CSS_PATH = "./demo_dock.css"

    def compose(self):
        yield SidebarContainer(id="sidebar-container")
        yield CalendarContainer(id="calendar-container")

    def on_category_tree_category_changed(self, event: CategoryTree.CategoryChanged):
        self.query_one(CalendarContainer).update_category(event.category)
        cat_tree = self.query_one(CategoryTree)

    def on_tag_tree_tag_changed(self, event):
        tag = event.tag
        if tag.endswith(" " + Emoji.LEFT_ARROW):
            tag = tag[:-2]
        if tag == "All Tags":
            tag = None
        self.query_one(CalendarContainer).update_tag(tag)

    def on_category_tree_custom_test_message(self, event):
        self.refresh()



import re

import typer
from rich import print
from rich.console import Console
from rich.markdown import Markdown
from rich.style import Style
from rich.text import Text
from rich.tree import Tree
from typing_extensions import Annotated

import girok.api.category as category_api
from girok.commands.category.util import (
    display_categories_tree,
    display_category_color_palette,
    get_next_category_color,
)
from girok.constants import (
    CATEGORY_COLOR_PALETTE,
    DEFAULT_CATEGORY_TEXT_COLOR,
    DisplayArrowType,
    DisplayBoxType,
    Emoji,
)
from girok.utils.display import arrow_print, center_print

app = typer.Typer(rich_markup_mode="rich")
console = Console()


def category_callback(ctx: typer.Context, param: typer.CallbackParam, value: str):
    if value is None:
        return None

    if ctx.command.name == "mvcat" and param.name == "new_parent_path" and value == "/":
        return value.rstrip("/")

    if not re.match("^([a-zA-Z0-9]+/)*[a-zA-Z0-9]+/?$", value):
        raise typer.BadParameter("[Invalid category path] Category path must be in 'xx/yy/zz format.'")

    if value.endswith("/"):
        value = value[:-1]

    if value == "none":
        raise typer.BadParameter("Sorry, 'none' is a reserved category name.")
    return value


@app.command(
    "showcat",
    help="[yellow]Show[/yellow] all pre-defined categories",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def show_categories():
    resp = category_api.get_all_categories()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Event Categories", DisplayBoxType.TITLE)
    root_categories: list[dict] = resp.body["rootCategories"]
    display_categories_tree(root_categories)


@app.command(
    "addcat",
    help="[yellow]Add[/yellow] a new category",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def add_category(
    category_path: Annotated[
        str,
        typer.Argument(
            ...,
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=category_callback,
        ),
    ],
    color: Annotated[str, typer.Option("-c", "--color", help="[yellow]Color[/yellow] for category")] = None,
):
    # Resolve color
    if color:
        if len(category_path.split("/")) != 1:
            arrow_print("You cannot specify non top-level category color", DisplayArrowType.ERROR)
            raise typer.Exit()

        if color not in CATEGORY_COLOR_PALETTE:
            arrow_print("Unsupported category color\n", DisplayArrowType.ERROR)
            display_category_color_palette()
            raise typer.Exit()
    else:
        # If color is not passed and top-level category, automatically assign the color from the palette
        if len(category_path.split("/")) == 1:
            color = get_next_category_color()

    # Create a new category
    resp = category_api.create_category(category_path, color)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    resp = category_api.get_all_categories()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Event Categories", DisplayBoxType.TITLE)
    root_categories: list[dict] = resp.body["rootCategories"]
    display_categories_tree(root_categories, category_path)


@app.command(
    "rmcat",
    help="[red]Remove[/red] a category",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def remove_category(
    category_path: Annotated[
        str,
        typer.Argument(
            ...,
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=category_callback,
        ),
    ],
    force_yes: Annotated[bool, typer.Option("-y", "--yes", help="[yellow]Ignore confirm message[/yellow]")] = False,
):
    if not force_yes:
        confirm_rm = typer.confirm(
            f"[WARNING] Are you sure to delete '{category_path}'?\nAll the subcategories and tasks will also be deleted."
        )
        if not confirm_rm:
            raise typer.Exit()

    resp = category_api.remove_category(category_path)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    resp = category_api.get_all_categories()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Event Categories", DisplayBoxType.TITLE)
    root_categories: list[dict] = resp.body["rootCategories"]
    display_categories_tree(root_categories, category_path)


@app.command(
    "upcat",
    help="[green]Rename[/green] a category",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def rename_category(
    category_path: Annotated[
        str,
        typer.Argument(
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=category_callback,
        ),
    ],
    new_name: Annotated[str, typer.Option("-n", "--name", help="[yellow]New category name[/yellow]")] = None,
    new_color: Annotated[str, typer.Option("-c", "--color", help="[yellow]New category color[/yellow]")] = None,
):
    if new_name is None and new_color is None:
        arrow_print("Please provide fields to update", DisplayArrowType.ERROR)
        raise typer.Exit()

    # Resolve color
    if new_color:
        if len(category_path.split("/")) != 1:
            arrow_print("You cannot update non top-level category color", DisplayArrowType.ERROR)
            raise typer.Exit()

        if new_color not in CATEGORY_COLOR_PALETTE:
            arrow_print("Unsupported category color\n", DisplayArrowType.ERROR)
            display_category_color_palette()
            raise typer.Exit()

    resp = category_api.update_category(category_path, new_name, new_color)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    resp = category_api.get_all_categories()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Event Categories", DisplayBoxType.TITLE)
    root_categories: list[dict] = resp.body["rootCategories"]
    name = new_name if new_name else category_path.split("/")[-1]
    new_path = "/".join(category_path.split("/")[:-1] + [name])
    display_categories_tree(root_categories, new_path)


@app.command(
    "mvcat",
    help="[yellow]Move[/yellow] a category to under category",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def move_category(
    path: Annotated[
        str,
        typer.Argument(
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=category_callback,
        ),
    ],
    new_parent_path: Annotated[
        str,
        typer.Argument(
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=category_callback,
        ),
    ],
):
    resp = category_api.move_category(path, new_parent_path)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    resp = category_api.get_all_categories()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Event Categories", DisplayBoxType.TITLE)
    root_categories: list[dict] = resp.body["rootCategories"]
    new_path = "/".join(new_parent_path.split("/") + [path.split("/")[-1]])
    display_categories_tree(root_categories, new_path)


@app.command(
    "colors",
    help="[yellow]Show[/yellow] category color palette",
    rich_help_panel=":file_folder: [bold yellow1]Category Commands[/bold yellow1]",
)
def show_category_color_palette():
    display_category_color_palette()



import calendar
from datetime import datetime, timedelta
from typing import List, Optional

import typer
from rich import print
from typing_extensions import Annotated

import girok.api.category as category_api
import girok.api.task as task_api
from girok.commands.task.callbacks import (
    allow_empty_category_callback,
    date_callback,
    datetime_callback,
    not_allow_empty_category_callback,
    priority_callback,
    tags_callback,
    valid_integer_callback,
)
from girok.commands.task.display import display_events_by_list, display_events_by_tree
from girok.commands.task.entity import Category, Event, EventDate, Repetition
from girok.commands.task.utils import decode_date_format, validate_start_end_window
from girok.constants import EVENT_IDS_CACHE_PATH, REPETITION_TYPE, DisplayBoxType
from girok.utils.display import center_print
from girok.utils.json_utils import read_json
from girok.utils.time import build_date_info, convert_date_obj_to_iso_date_str

app = typer.Typer(rich_markup_mode="rich")


@app.command(
    "addtask",
    help="[yellow]Add[/yellow] a new task",
    rich_help_panel=":fire: [bold yellow1]Task Commands[/bold yellow1]",
)
def add_task(
    name: Annotated[str, typer.Argument(help="Task name")],
    start_datetime: Annotated[
        str, typer.Option("-d", "--date", help="[yellow]Task start datetime[/yellow]", callback=datetime_callback)
    ],
    end_datetime: Annotated[
        Optional[str],
        typer.Option("-e", "--end", help="[yellow]Task end datetime[/yellow]", callback=datetime_callback),
    ] = None,
    repetition: Annotated[
        Optional[str],
        typer.Option(
            "-r",
            "--repetition",
            help="[yellow]Task repetition type. One of 'daily', 'weekly', 'monthly', 'yearly'[/yellow]",
        ),
    ] = None,
    category_path: Annotated[
        Optional[str],
        typer.Option(
            "-c",
            "--category",
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=allow_empty_category_callback,
        ),
    ] = None,
    tags: Annotated[
        Optional[str],
        typer.Option(
            "-t",
            "--tag",
            help="[yellow]Tags[/yellow]. Multiple tags must be provided in 'A,B,C' format.",
            callback=tags_callback,
        ),
    ] = None,
    priority: Annotated[
        Optional[str], typer.Option("-p", "--priority", help="[yellow]Priority[/yellow]", callback=priority_callback)
    ] = None,
    memo: Annotated[Optional[str], typer.Option("-m", "--memo", help="[yellow]Memo[/yellow]")] = None,
):
    """
    Validate time combination. The possible combinations are:
    1. start_date
    2. start_date, start_time
    3. start_date, end_date
    4. start_date, start_time, end_date, end_time
    """
    start_date, start_time = start_datetime
    end_date, end_time = None, None
    if end_datetime:
        end_date, end_time = end_datetime

    valid, err_msg = validate_start_end_window(start_date, start_time, end_date, end_time)
    if not valid:
        raise typer.BadParameter(err_msg)

    # Convert tags to list
    if tags:
        tags = tags.split(",")

    # Validate repetition
    repetition_type = None
    repetition_end_date = None
    if repetition:
        # Repetition is only allowed for single-day event
        if (start_date and end_date) and (start_date != end_date):
            raise typer.BadParameter("Repetition is only allowed for single-day event")

        if "@" not in repetition:  # daily
            if repetition not in REPETITION_TYPE:
                raise typer.BadParameter("Repetition type must be one of 'daily', 'weekly', 'monthly', 'yearly'")
            repetition_type = repetition
        else:  # daily@5/14
            repetition_items = repetition.split("@")
            if len(repetition_items) != 2:
                raise typer.BadParameter("Invalid repetition input format")

            repetition_type, repetition_end_date_str = repetition_items
            if repetition_type not in REPETITION_TYPE:
                raise typer.BadParameter("Repetition type must be one of 'daily', 'weekly', 'monthly', 'yearly'")

            valid, iso_date_str = decode_date_format(repetition_end_date_str)
            if not valid:
                raise typer.BadParameter("Invalid repetition end date format")

            repetition_end_date = iso_date_str

            # Repetition end date must be greater than start_date
            repetition_end_date_obj = datetime.strptime(repetition_end_date, "%Y-%m-%d").date()
            start_date_obj = datetime.strptime(start_date, "%Y-%m-%d").date()
            if repetition_end_date_obj <= start_date_obj:
                raise typer.BadParameter("Repetition end date must be greater than start date")

        repetition_type = REPETITION_TYPE[repetition_type]

    resp = task_api.create_task(
        name=name,
        start_date=start_date,
        start_time=start_time,
        end_date=end_date,
        end_time=end_time,
        repetition_type=repetition_type,
        repetition_end_date=repetition_end_date,
        category_path=category_path,
        tags=tags,
        priority=priority,
        memo=memo,
    )
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    # Display Tasks
    created_event_id = resp.body["eventId"]
    resp = task_api.get_all_tasks()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    events = resp.body["events"]
    event_entities = map_to_event_entities(events)
    display_events_by_list(event_entities, highlight_event_id=created_event_id, highlight_action="highlight")


@app.command(
    "showtask",
    help="[yellow]View[/yellow] tasks with options",
    rich_help_panel=":fire: [bold yellow1]Task Commands[/bold yellow1]",
)
def showtask(
    category_path: Annotated[
        Optional[str],
        typer.Option(
            "-c",
            "--category",
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=not_allow_empty_category_callback,
        ),
    ] = None,
    tags: Annotated[
        Optional[str],
        typer.Option(
            "-t",
            "--tag",
            help="[yellow]Tags[/yellow]. Multiple tags must be provided in 'A,B,C' format.",
            callback=tags_callback,
        ),
    ] = None,
    priority: Annotated[
        Optional[str], typer.Option("-p", "--priority", help="[yellow]Priority[/yellow]", callback=priority_callback)
    ] = None,
    exact_date: Annotated[
        Optional[str], typer.Option("-e", "--exact", help="[yellow]Exact Deadline[/yellow]", callback=date_callback)
    ] = None,
    within_days: Annotated[
        Optional[int],
        typer.Option(
            "-d",
            "--day",
            help="Show tasks due [yellow]within the specified days[/yellow]",
            callback=valid_integer_callback,
        ),
    ] = None,
    within_weeks: Annotated[
        Optional[int],
        typer.Option(
            "-w",
            "--week",
            help="Show tasks due [yellow]within the specified weeks[/yellow]",
            callback=valid_integer_callback,
        ),
    ] = None,
    within_months: Annotated[
        Optional[int],
        typer.Option(
            "-m",
            "--month",
            help="Show tasks due [yellow]within the specified months[/yellow]",
            callback=valid_integer_callback,
        ),
    ] = None,
    within_this_week: Annotated[
        Optional[bool],
        typer.Option(
            "-tw",
            "--this-week",
            help="Show tasks due [yellow]within this week[/yellow]",
        ),
    ] = None,
    within_next_week: Annotated[
        Optional[bool],
        typer.Option(
            "-nw",
            "--next-week",
            help="Show tasks due [yellow]within next week[/yellow]",
        ),
    ] = None,
    within_this_month: Annotated[
        Optional[bool],
        typer.Option(
            "-tm",
            "--this-month",
            help="Show tasks due [yellow]within this month[/yellow]",
        ),
    ] = None,
    within_next_month: Annotated[
        Optional[bool],
        typer.Option(
            "-nm",
            "--next-month",
            help="Show tasks due [yellow]within next month[/yellow]",
        ),
    ] = None,
    today: Annotated[
        Optional[bool],
        typer.Option(
            "-tdy",
            "--today",
            help="Show tasks due [yellow]today[/yellow]",
        ),
    ] = None,
    tomorrow: Annotated[
        Optional[bool],
        typer.Option(
            "-tmr",
            "--tomorrow",
            help="Show tasks due [yellow]tomorrow[/yellow]",
        ),
    ] = None,
    urgent: Annotated[
        Optional[bool], typer.Option("-u", "--urgent", help="Show [yellow]urgent[/yellow] tasks (due within 3 days)")
    ] = None,
    tree_view: Annotated[
        Optional[bool],
        typer.Option(
            '--tree',
            help="[yellow]Show tasks in a tree view[/yellow]"
        )
    ] = None
):
    # Resolve start_date and end_date
    """
    1. exact_date -> [start, end]
    2. within_days -> [-inf, end]
    3. within_weeks -> [-inf, end]
    4. within_months -> [-inf, month]
    5. within_this_week -> [-inf, end]
    6. within_next_week -> [-inf, end]
    7. within_this_month -> [-inf, end]
    8. within_next_month -> [-inf, end]
    9. today -> [start, end]
    10. tomorrow -> [start, end]
    11. urgent -> [today, end]
    -> else -> [-inf, inf]
    """
    date_options_cnt = sum(
        [
            opt is not None
            for opt in [
                exact_date,
                within_days,
                within_weeks,
                within_months,
                today,
                tomorrow,
                within_this_week,
                within_next_week,
                within_this_month,
                within_next_month,
            ]
        ]
    )
    if date_options_cnt > 1:
        raise typer.BadParameter("You can specify only one date option.")

    start_date, end_date = "2000-01-01", convert_date_obj_to_iso_date_str(datetime.now() + timedelta(days=365))
    if exact_date:
        start_date, end_date = exact_date, exact_date
    if within_days:
        end_date = convert_date_obj_to_iso_date_str(datetime.now() + timedelta(days=within_days - 1))
    if within_weeks:
        end_date = convert_date_obj_to_iso_date_str(datetime.now() + timedelta(days=7 * within_weeks))
    if within_months:
        end_date = convert_date_obj_to_iso_date_str(datetime.now() + timedelta(days=30 * within_months))
    if within_this_week:
        delta = 6 - datetime.today().weekday()
        end_date = convert_date_obj_to_iso_date_str(datetime.today() + timedelta(days=delta))
    if within_next_week:
        next_week = datetime.today() + timedelta(days=7)
        delta = 6 - next_week.weekday()
        end_date = convert_date_obj_to_iso_date_str(next_week + timedelta(days=delta))
    if within_this_month:
        today_date = datetime.today()
        _, last_day = calendar.monthrange(today_date.year, today_date.month)
        end_date = convert_date_obj_to_iso_date_str(datetime(today_date.year, today_date.month, last_day))
    if within_next_month:
        today_date = datetime.today()
        if today_date.month == 12:
            first_day_next_month = datetime(today_date.year + 1, 1, 1)
        else:
            first_day_next_month = datetime(today_date.year, today_date.month + 1, 1)
        _, last_day = calendar.monthrange(first_day_next_month.year, first_day_next_month.month)
        end_date = convert_date_obj_to_iso_date_str(
            datetime(first_day_next_month.year, first_day_next_month.month, last_day)
        )
    if today:
        start_date, end_date = convert_date_obj_to_iso_date_str(datetime.now()), convert_date_obj_to_iso_date_str(
            datetime.now()
        )
    if tomorrow:
        start_date, end_date = convert_date_obj_to_iso_date_str(
            datetime.now() + timedelta(1)
        ), convert_date_obj_to_iso_date_str(datetime.now() + timedelta(1))
    if urgent:
        start_date, end_date = convert_date_obj_to_iso_date_str(datetime.now()), convert_date_obj_to_iso_date_str(
            datetime.now() + timedelta(days=2)
        )

    # Resolve category id
    category_id = None
    if category_path:
        resp = category_api.get_category_id_by_path(category_path.split("/"))
        if not resp.is_success:
            center_print(resp.error_message, DisplayBoxType.ERROR)
            raise typer.Exit()
        category_id = resp.body["categoryId"]

    if tags:
        tags = tags.split("/")

    resp = task_api.get_all_tasks(
        start_date=start_date, end_date=end_date, category_id=category_id, priority=priority, tags=tags
    )
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    events = resp.body["events"]

    # Display Events
    event_entities = map_to_event_entities(events)
    center_print(build_date_info(datetime.now()), DisplayBoxType.TITLE)
    if tree_view:
        resp = category_api.get_all_categories()
        if not resp.is_success:
            center_print("Unable to fetch categories", DisplayBoxType.ERROR)
            raise typer.Exit()
        categories = resp.body['rootCategories']
        display_events_by_tree(categories, event_entities)
    else:
        display_events_by_list(event_entities)


@app.command(
    "done",
    help="[red]Delete[/red] a task",
    rich_help_panel=":fire: [bold yellow1]Task Commands[/bold yellow1]",
)
def remove_event(
    event_id: Annotated[int, typer.Argument(help="[yellow]Task ID[/yellow] to be deleted")],
    force: Annotated[Optional[bool], typer.Option("-y", "--yes", help="Don't show the confirmation message")] = False,
):
    try:
        cache = read_json(EVENT_IDS_CACHE_PATH)
    except:
        center_print("First type 'showtask' command to retrieve task ids", DisplayBoxType.ERROR)
        raise typer.Exit()

    if str(event_id) not in cache:
        center_print(f"Task id {event_id} is not found. Please enter 'showtask' command to view task ids.", DisplayBoxType.ERROR)
        raise typer.Exit()
    cached_event = cache[str(event_id)]

    target_event_id = cached_event["id"]
    target_event_name = cached_event["name"]

    if not force:
        done_confirm = typer.confirm(f"Are you sure to delete task [{target_event_name}]?")
        if not done_confirm:
            raise typer.Exit()

    # Display events
    resp = task_api.get_all_tasks()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    events = resp.body["events"]
    event_entities = map_to_event_entities(events)

    # Remove events
    resp = task_api.remove_event(target_event_id)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    center_print("Task was successfully deleted!", DisplayBoxType.SUCCESS)
    display_events_by_list(event_entities, highlight_event_id=target_event_id, highlight_action="delete")


def map_to_event_entities(events: List[dict]) -> List[Event]:
    event_entities: List[Event] = []
    for event in events:
        event_entity = Event(
            id=event["id"],
            name=event["name"],
            color_str=event["color"],
            tags=event["tags"],
            priority=event["priority"],
            memo=event["memo"],
            event_date=EventDate(
                start_date=event["eventDate"]["startDate"],
                start_time=event["eventDate"]["startTime"],
                end_date=event["eventDate"]["endDate"],
                end_time=event["eventDate"]["endTime"],
            ),
            repetition=Repetition(
                repetition_type=event["repetition"]["repetitionType"],
                repetition_end_date=event["repetition"]["repetitionEndDate"],
            ),
            category_path=[Category(id=c["categoryId"], name=c["categoryName"]) for c in event["categoryPath"]],
        )

        event_entities.append(event_entity)

    def sort_key(event: Event):
        # Handle None values for start_time and end_time by replacing them with "00:00"
        start_time = event.event_date.start_time if event.event_date.start_time else "00:00"
        end_time = event.event_date.end_time if event.event_date.end_time else "00:00"
        # Handle None for end_date by using start_date
        end_date = event.event_date.end_date if event.event_date.end_date else event.event_date.start_date

        return (event.event_date.start_date, start_time, end_date, end_time)

    event_entities.sort(key=sort_key)
    return event_entities


@app.command(
    "uptask",
    help="[yellow]Update[/yellow] a new task",
    rich_help_panel=":fire: [bold yellow1]Task Commands[/bold yellow1]",
)
def update_task(
    event_id: Annotated[
        int,
        typer.Argument(
            help="[yellow]Task ID[/yellow]"
        )
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            '-n', '--name',
            help="Task name"
        )
    ] = None,
    start_datetime: Annotated[
        Optional[str],
        typer.Option("-d", "--date", help="[yellow]Task start datetime[/yellow]", callback=datetime_callback)
    ] = None,
    end_datetime: Annotated[
        Optional[str],
        typer.Option("-e", "--end", help="[yellow]Task end datetime[/yellow]", callback=datetime_callback),
    ] = None,
    repetition: Annotated[
        Optional[str],
        typer.Option(
            "-r",
            "--repetition",
            help="[yellow]Task repetition type. One of 'daily', 'weekly', 'monthly', 'yearly'[/yellow]",
        ),
    ] = None,
    category_path: Annotated[
        Optional[str],
        typer.Option(
            "-c",
            "--category",
            help="[yellow]Category path - xx/yy/zz..[/yellow]",
            callback=allow_empty_category_callback,
        ),
    ] = None,
    tags: Annotated[
        Optional[str],
        typer.Option(
            "-t",
            "--tag",
            help="[yellow]Tags[/yellow]. Multiple tags must be provided in 'A,B,C' format.",
            callback=tags_callback,
        ),
    ] = None,
    priority: Annotated[
        Optional[str], typer.Option("-p", "--priority", help="[yellow]Priority[/yellow]", callback=priority_callback)
    ] = None,
    memo: Annotated[Optional[str], typer.Option("-m", "--memo", help="[yellow]Memo[/yellow]")] = None,
):
    try:
        cache = read_json(EVENT_IDS_CACHE_PATH)
    except:
        center_print("First type 'showtask' command to retrieve task ids", DisplayBoxType.ERROR)
        raise typer.Exit()

    if str(event_id) not in cache:
        center_print(f"Task id {event_id} is not found. Please enter 'showtask' command to view task ids.", DisplayBoxType.ERROR)
        raise typer.Exit()
    cached_event = cache[str(event_id)]

    target_event_id = cached_event["id"]

    """
    Validate time combination. The possible combinations are:
    1. start_date
    2. start_date, start_time
    3. start_date, end_date
    4. start_date, start_time, end_date, end_time
    """
    start_date, start_time = None, None
    if start_datetime:
        start_date, start_time = start_datetime

    end_date, end_time = None, None
    if end_datetime:
        end_date, end_time = end_datetime
    
    if start_date or start_time or end_date or end_time:
        valid, err_msg = validate_start_end_window(start_date, start_time, end_date, end_time)
        if not valid:
            raise typer.BadParameter(err_msg)

    # Convert tags to list
    if tags:
        tags = tags.split(",")

    # Validate repetition
    repetition_type = None
    repetition_end_date = None
    if repetition:
        # Repetition is only allowed for single-day event
        if (start_date and end_date) and (start_date != end_date):
            raise typer.BadParameter("Repetition is only allowed for single-day event")

        if "@" not in repetition:  # daily
            if repetition not in REPETITION_TYPE:
                raise typer.BadParameter("Repetition type must be one of 'daily', 'weekly', 'monthly', 'yearly'")
            repetition_type = repetition
        else:  # daily@5/14
            repetition_items = repetition.split("@")
            if len(repetition_items) != 2:
                raise typer.BadParameter("Invalid repetition input format")

            repetition_type, repetition_end_date_str = repetition_items
            if repetition_type not in REPETITION_TYPE:
                raise typer.BadParameter("Repetition type must be one of 'daily', 'weekly', 'monthly', 'yearly'")

            valid, iso_date_str = decode_date_format(repetition_end_date_str)
            if not valid:
                raise typer.BadParameter("Invalid repetition end date format")

            repetition_end_date = iso_date_str

            # Repetition end date must be greater than start_date
            repetition_end_date_obj = datetime.strptime(repetition_end_date, "%Y-%m-%d").date()
            start_date_obj = datetime.strptime(start_date, "%Y-%m-%d").date()
            if repetition_end_date_obj <= start_date_obj:
                raise typer.BadParameter("Repetition end date must be greater than start date")

        repetition_type = REPETITION_TYPE[repetition_type]

    
    # Get the target event
    resp = task_api.get_single_event(target_event_id)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()
    event = resp.body

    if category_path is None:
        category_path = '/'.join([c['categoryName'] for c in event['categoryPath']])    

    resp = task_api.update_task(
        event_id=target_event_id,
        name=name if name else event['name'],
        start_date=start_date if start_date else event['eventDate']['startDate'],
        start_time=start_time if start_time else event['eventDate']['startTime'],
        end_date=end_date if end_date else event['eventDate']['endDate'],
        end_time=end_time if end_time else event['eventDate']['endTime'],
        repetition_type=repetition_type if repetition_type else event['repetition']['repetitionType'],
        repetition_end_date=repetition_end_date if repetition_end_date else event['repetition']['repetitionEndDate'],
        category_path=category_path,
        tags=tags if tags else event['tags'],
        priority=priority if priority else event['priority'],
        memo=memo if memo else event['memo'],
    )
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    # Display Tasks
    resp = task_api.get_all_tasks()
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    events = resp.body["events"]
    event_entities = map_to_event_entities(events)
    display_events_by_list(event_entities, highlight_event_id=target_event_id, highlight_action="highlight")



from typing import Optional
from urllib.parse import urljoin

import requests
from requests import HTTPError

from girok.api.entity import APIResponse
from girok.config.auth_handler import AuthHandler
from girok.constants import BASE_URL


def get_all_categories() -> APIResponse:
    access_token = AuthHandler.get_access_token()
    resp = requests.get(
        url=urljoin(BASE_URL, "categories"),
        headers={"Authorization": "Bearer " + access_token},
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError as e:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to get categories"

        return APIResponse(is_success=False, error_message=error_message)


def create_category(category_path: str, color: str) -> APIResponse:
    access_token = AuthHandler.get_access_token()

    category_path_list = category_path.split("/")
    new_category_name = category_path_list[-1]

    # Resolve parent category's id
    parent_category_id_resp = get_category_id_by_path(category_path_list[:-1])
    if not parent_category_id_resp.is_success:
        return APIResponse(is_success=False, error_message=parent_category_id_resp.error_message)

    parent_category_id = parent_category_id_resp.body["categoryId"]
    resp = requests.post(
        url=urljoin(BASE_URL, "categories"),
        headers={"Authorization": "Bearer " + access_token},
        json={"parentId": parent_category_id, "name": new_category_name, "color": color},
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_code = error_body["errorCode"]
            error_message = error_body["message"]

            if error_code == "DUPLICATE_CATEGORY":
                parent_category_path_str = (
                    "/" if not category_path_list[:-1] else "/".join(category_path_list[:-1]) + "/"
                )
                error_message = f"Duplicate Category: '{parent_category_path_str}' already has '{new_category_name}'"
        except:
            error_message = "Failed to create a new category"

        return APIResponse(is_success=False, error_message=error_message)


def remove_category(category_path: str) -> APIResponse:
    access_token = AuthHandler.get_access_token()

    category_path_list = category_path.split("/")
    category_id_resp = get_category_id_by_path(category_path_list)
    if not category_id_resp.is_success:
        return APIResponse(is_success=False, error_message=category_id_resp.error_message)

    category_id = category_id_resp.body["categoryId"]
    resp = requests.delete(
        url=urljoin(BASE_URL, f"categories/{category_id}"),
        headers={"Authorization": "Bearer " + access_token},
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True)
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to remove a category"

        return APIResponse(is_success=False, error_message=error_message)


def update_category(category_path: str, new_name: Optional[str] = None, new_color: Optional[str] = None) -> APIResponse:
    access_token = AuthHandler.get_access_token()

    category_path_list = category_path.split("/")
    category_id_resp = get_category_id_by_path(category_path_list)
    if not category_id_resp.is_success:
        return category_id_resp

    category_id = category_id_resp.body["categoryId"]
    body = {}
    if new_name:
        body["newName"] = new_name
    if new_color:
        body["color"] = new_color

    resp = requests.patch(
        url=urljoin(BASE_URL, f"categories/{category_id}"),
        headers={"Authorization": "Bearer " + access_token},
        json=body,
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True)
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to rename a category"

        return APIResponse(is_success=False, error_message=error_message)


def move_category(path: str, new_parent_path: str) -> APIResponse:
    # girok mvcat A/B/C D/E
    access_token = AuthHandler.get_access_token()

    path_list = path.split("/")
    new_parent_path_list = new_parent_path.split("/") if new_parent_path else []

    # 1. Get the category id
    resp = get_category_id_by_path(path_list)
    if not resp.is_success:
        return resp
    category_id = resp.body["categoryId"]

    # 2. Get target parent's id
    resp = get_category_id_by_path(new_parent_path_list)
    if not resp.is_success:
        return resp
    new_parent_category_id = resp.body["categoryId"]

    resp = requests.patch(
        url=urljoin(BASE_URL, f"categories/{category_id}/parent"),
        headers={"Authorization": "Bearer " + access_token},
        json={"newParentId": new_parent_category_id},
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True)
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to move a category"
        return APIResponse(is_success=False, error_message=error_message)


def get_category_id_by_path(path_list: list[str]) -> APIResponse:
    if len(path_list) == 0:
        return APIResponse(is_success=True, body={"categoryId": None})

    access_token = AuthHandler.get_access_token()
    resp = requests.get(
        url=urljoin(BASE_URL, "categories/id-by-path"),
        headers={"Authorization": "Bearer " + access_token},
        params={"path": path_list},
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = f"Failed to get a category id of '{path_list}'"
        return APIResponse(is_success=False, error_message=error_message)



import calendar
from datetime import datetime, timedelta


def convert_iso_date_str_to_date_obj(date_str: str) -> datetime:
    return datetime.strptime(date_str, "%Y-%m-%d")


def convert_date_obj_to_iso_date_str(date_obj: datetime) -> str:
    return date_obj.strftime("%Y-%m-%d")


def get_end_of_current_month_date() -> datetime:
    today = datetime.today()
    _, last_day = calendar.monthrange(today.year, today.month)
    end_date = convert_date_obj_to_iso_date_str(datetime(today.year, today.month, last_day))


def get_day_offset(dt1: datetime, dt2: datetime):
    dt1 = datetime.strptime(f"{dt1.year}-{dt1.month}-{dt1.day}", "%Y-%m-%d")
    dt2 = datetime.strptime(f"{dt2.year}-{dt2.month}-{dt2.day}", "%Y-%m-%d")
    diff = dt2 - dt1
    return diff.days


def build_date_info(dt: datetime):
    year, month, day = dt.year, dt.month, dt.day
    h, m, s = str(dt.time()).split(":")
    month_name = calendar.month_name[month]
    weekday_name = calendar.day_name[dt.weekday()]
    return f"{weekday_name}, {month_name} {day}, {year}"


def get_year_and_month_by_month_offset(month_offset: int, year=None, month=None):
    if year is None and month is None:
        now = datetime.now()
        year, month = now.year, now.month
    new_months = year * 12 + month + month_offset
    new_year = new_months // 12
    new_month = new_months % 12
    if new_month == 0:  # (1 - 1) % 12 = 0.
        new_year -= 1
        new_month = 12
    return new_year, new_month


import os

import typer
from rich import print

import girok.commands.auth.command as auth_command
import girok.commands.category.command as category_command
import girok.commands.task.command as task_command
import girok.commands.calendar.command as calendar_command
from girok.config.auth_handler import AuthHandler
from girok.constants import VERSION, CommandName

app = typer.Typer(
    rich_markup_mode="rich",
    help="Enter [red]girok <command name> --help[/red] to see more detailed documentations of commands!",
)

app.registered_commands.extend(auth_command.app.registered_commands)
app.registered_commands.extend(category_command.app.registered_commands)
app.registered_commands.extend(task_command.app.registered_commands)
app.registered_commands.extend(calendar_command.app.registered_commands)


@app.command("version")
def version():
    print(VERSION)


@app.callback()
def pre_command_callback(ctx: typer.Context):
    # Get the executed command name
    cmd = ctx.invoked_subcommand

    # Set up application directory and config.json
    AuthHandler.init()

    # Utility commands
    if cmd in [CommandName.VERSION, CommandName.COLORS]:
        return

    """
    Login, Register -> login 되면 X
    나머지 -> login 필수
    """
    is_logged_in = AuthHandler.is_logged_in()

    # Logout required commands
    if cmd in [CommandName.REGISTER, CommandName.LOGIN]:
        if is_logged_in:
            print("[red]You're already logged in. Please log out and try again.[/red]")
            raise typer.Exit()
    # Login required commands
    else:
        if not is_logged_in:
            print("[red]You're not logged in. Please login with [/red][yellow]girok login[/yellow].")
            raise typer.Exit()


if __name__ == "__main__":
    app()



import re
from datetime import date, datetime, timedelta
from typing import Callable, List, Tuple

import typer

from girok.constants import EVENT_IDS_CACHE_PATH
from girok.utils.json_utils import read_json, write_json


def decode_date_format(date_str: str) -> Tuple[bool, str]:
    """Decode date input into ISO date format, 'yyyy-mm-dd'.

    Args:
        date_str (str): The input date string for `-d` command option.

    Returns:
        Tuple[bool, str]: is_success, ISO date format('yyyy-mm-dd')
    """
    decode_functions: List[Callable[[str], Tuple[bool, str]]] = [
        decode_direct_date_format,
        decode_this_week_date_format,
        decode_next_week_date_format,
        decode_after_date_format,
        decode_today_date_format,
        decode_tomorrow_date_format,
    ]

    for decode_func in decode_functions:
        valid, iso_date_str = decode_func(date_str)
        if valid:
            return True, iso_date_str

    return False, None


def decode_direct_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: 2024/05/26, 2024/5/26, 05/26, 5/26
    """
    is_matched = bool(re.match("^([0-9]){4}/([0-9]){1,2}/([0-9]){1,2}|([0-9]){1,2}/([0-9]){1,2}$", date_str))

    if not is_matched:
        return False, None

    date_items = [int(i) for i in date_str.split("/")]
    year, month, day = datetime.now().year, None, None
    if len(date_items) == 3:
        year, month, day = date_items
    elif len(date_items) == 2:
        month, day = date_items
    else:
        raise ValueError("Invalid date_str format")

    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_this_week_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: t1, t2, ..., t7
    """
    if date_str not in ["t1", "t2", "t3", "t4", "t5", "t6", "t7"]:
        return False, None

    today = date.today()
    monday = today - timedelta(days=today.weekday())

    delta = int(date_str[1]) - 1
    target = monday + timedelta(days=delta)
    year, month, day = target.year, target.month, target.day
    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_next_week_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: n1, n2, ..., n7
    """
    if date_str not in ["n1", "n2", "n3", "n4", "n5", "n6", "n7"]:
        return False, None

    today = date.today()
    next_monday = today - timedelta(days=today.weekday()) + timedelta(days=7)
    delta = int(date_str[1]) - 1
    target = next_monday + timedelta(days=delta)
    year, month, day = target.year, target.month, target.day
    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_after_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: a1, a10, a50, ...
    """
    is_matched = bool(re.match("^a[1-9]\d*$", date_str))
    if not is_matched:
        return False, None
    delta = int(date_str[1:])
    target = date.today() + timedelta(days=delta)
    year, month, day = target.year, target.month, target.day
    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_today_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: tdy, today
    """
    if date_str not in ["tdy", "today"]:
        return False, None

    target = date.today()
    year, month, day = target.year, target.month, target.day
    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_tomorrow_date_format(date_str: str) -> Tuple[bool, str]:
    """
    date_str: tmr, tomorrow
    """
    if date_str not in ["tmr", "tomorrow"]:
        return False, None

    target = date.today() + timedelta(days=1)
    year, month, day = target.year, target.month, target.day
    try:
        iso_date_str = to_iso_date_str(year, month, day)
        return True, iso_date_str
    except ValueError:
        return False, None


def decode_time_format(time_str: str) -> Tuple[bool, str]:
    is_matched = bool(re.match("^[01][0-9]|2[0-3]:[0-5][0-9]$", time_str))
    if not is_matched:
        return False, None

    return True, time_str + ":00"


def to_iso_date_str(year: int, month: int, day: int) -> str:
    return date(year, month, day).strftime("%Y-%m-%d")


def validate_start_end_window(start_date: str, start_time: str, end_date: str, end_time: str) -> bool:
    if start_date and not start_time and not end_date and not end_time:
        # start_date
        return True, None
    elif start_date and start_time and not end_date and not end_time:
        # start_date, start_time
        return True, None
    elif start_date and not start_time and end_date and not end_time:
        # start_date, end_date
        start = datetime.strptime(start_date, "%Y-%m-%d").date()
        end = datetime.strptime(end_date, "%Y-%m-%d").date()
        if end <= start:
            return False, "End date must be greater than start date"
        return True, None
    elif start_date and start_time and end_date and end_time:
        # start_date, start_time, end_date, end_time
        start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
        start_time = datetime.strptime(start_time, "%H:%M:%S").time()
        start = datetime.combine(start_date, start_time)

        end_date = datetime.strptime(end_date, "%Y-%m-%d").date()
        end_time = datetime.strptime(end_time, "%H:%M:%S").time()
        end = datetime.combine(end_date, end_time)

        if end <= start:
            return False, "End date must be greater than start date"
        return True, None
    else:
        return (
            False,
            "Only allowed combinations are (1)start_date, (2)start_date, start_time, (3)start_date, end_date, (4)start_date, start_time, end_date, end_time",
        )


def cache_event_ids(event_ids_cache: dict):
    write_json(EVENT_IDS_CACHE_PATH, event_ids_cache)


def get_event_ids_cache():
    try:
        cache = read_json(EVENT_IDS_CACHE_PATH)
    except:
        raise typer.Exit()
    return cache



from typing import List, Optional
from urllib.parse import urljoin

import requests
from requests import HTTPError

from girok.api.category import get_category_id_by_path
from girok.api.entity import APIResponse
from girok.config.auth_handler import AuthHandler
from girok.constants import BASE_URL


def create_task(
    name: str,
    start_date: str,
    start_time: Optional[str],
    end_date: Optional[str],
    end_time: Optional[str],
    repetition_type: Optional[str],
    repetition_end_date: Optional[str],
    category_path: Optional[str],
    tags: Optional[List[str]],
    priority: Optional[str],
    memo: Optional[str],
) -> APIResponse:
    access_token = AuthHandler.get_access_token()

    # Resolve target category id
    category_id = None
    if category_path is not None:
        category_path_list = category_path.split("/")
        if category_path == '':
            category_path_list = []
        resp = get_category_id_by_path(category_path_list)
        if not resp.is_success:
            return resp
        category_id = resp.body["categoryId"]
    request_body = {
        "categoryId": category_id,
        "name": name,
        "eventDate": {"startDate": start_date, "startTime": start_time, "endDate": end_date, "endTime": end_time},
        "repetition": {"repetitionType": repetition_type, "repetitionEndDate": repetition_end_date},
        "tags": tags,
        "priority": priority,
        "memo": memo,
    }

    resp = requests.post(
        url=urljoin(BASE_URL, "events"), headers={"Authorization": "Bearer " + access_token}, json=request_body
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to create a new task"

        return APIResponse(is_success=False, error_message=error_message)


def update_task(
    event_id: int,
    name: str,
    start_date: str,
    start_time: Optional[str],
    end_date: Optional[str],
    end_time: Optional[str],
    repetition_type: Optional[str],
    repetition_end_date: Optional[str],
    category_path: Optional[str],
    tags: Optional[List[str]],
    priority: Optional[str],
    memo: Optional[str],
) -> APIResponse:
    access_token = AuthHandler.get_access_token()

    # Resolve target category id
    category_id = None
    if category_path is not None:
        category_path_list = category_path.split("/")
        if category_path == '':
            category_path_list = []
        resp = get_category_id_by_path(category_path_list)
        if not resp.is_success:
            return resp
        category_id = resp.body["categoryId"]

    request_body = {
        "categoryId": category_id,
        "name": name,
        "eventDate": {"startDate": start_date, "startTime": start_time, "endDate": end_date, "endTime": end_time},
        "repetition": {"repetitionType": repetition_type, "repetitionEndDate": repetition_end_date},
        "tags": tags,
        "priority": priority,
        "memo": memo,
    }

    resp = requests.put(
        url=urljoin(BASE_URL, f"events/{event_id}"),
        headers={"Authorization": "Bearer " + access_token},
        json=request_body
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True)
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to create a new task"

        return APIResponse(is_success=False, error_message=error_message)



def get_single_event(event_id: int) -> APIResponse:
    access_token = AuthHandler.get_access_token()
    resp = requests.get(
        url=urljoin(BASE_URL, f"events/{event_id}"),
        headers={"Authorization": "Bearer " + access_token}
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to retrieve a task"

        return APIResponse(is_success=False, error_message=error_message)


def get_all_tasks(
    start_date: Optional[str] = "2000-01-01",
    end_date: Optional[str] = "2050-01-01",
    category_id: Optional[int] = None,
    priority: Optional[str] = None,
    tags: Optional[List[str]] = None,
    fetch_children: bool = False
):
    params = {
        "startDate": start_date,
        "endDate": end_date,
        "categoryId": category_id,
        "priority": priority,
        "tags": tags,
        "fetchCategoryChildren": fetch_children
    }

    access_token = AuthHandler.get_access_token()
    resp = requests.get(
        url=urljoin(BASE_URL, "events"), headers={"Authorization": "Bearer " + access_token}, params=params
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to retrieve tasks"

        return APIResponse(is_success=False, error_message=error_message)


def remove_event(event_id: int):
    access_token = AuthHandler.get_access_token()
    resp = requests.delete(
        url=urljoin(BASE_URL, f"events/{event_id}"), headers={"Authorization": "Bearer " + access_token}
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True)
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to retrieve tasks"

        return APIResponse(is_success=False, error_message=error_message)


def get_all_tags():
    access_token = AuthHandler.get_access_token()
    resp = requests.get(
        url=urljoin(BASE_URL, "tags"),
        headers={"Authorization": "Bearer " + access_token}
    )

    try:
        resp.raise_for_status()
        return APIResponse(is_success=True, body=resp.json())
    except HTTPError:
        try:
            error_body = resp.json()
            error_message = error_body["message"]
        except:
            error_message = "Failed to retrieve tags"

        return APIResponse(is_success=False, error_message=error_message)


import shutil
import time

import typer
from rich import print
from rich.console import Console
from rich.live import Live
from rich.progress import Progress
from rich.style import Style
from rich.text import Text

import girok.api.auth as auth_api
from girok.config.auth_handler import AuthHandler
from girok.constants import DisplayBoxType
from girok.utils.display import center_print

app = typer.Typer(rich_markup_mode="rich")
console = Console()


@app.command(
    "register",
    help="[green]Register[/green] a new account",
    rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]",
)
def register():
    print(":star: [yellow]Welcome to girok![/yellow] :star:\n")

    is_register = typer.confirm("> Do you want to register a new account?")
    if not is_register:
        raise typer.Exit()

    # Send email verification code
    email = typer.prompt("\n> Enter email address")
    resp = auth_api.send_verification_code(email=email)
    if not resp.is_success:
        center_print(text=resp.error_message, text_type=DisplayBoxType.ERROR)
        raise typer.Exit()

    # Verify email verification code
    print("\n[yellow]Check your inbox to verify your email address.[/yellow]")
    print(
        "[yellow]If you don't find it in your inbox, please check[/yellow] [red]spam/junk[/red] [yellow]inbox[/yellow]."
    )
    retry_count = 3
    while retry_count > 0:
        verification_code = typer.prompt(f"> Enter verification code [{retry_count} tries left]")
        retry_count -= 1
        resp = auth_api.verify_verification_code(email=email, verification_code=verification_code)

        if resp.is_success:
            break
        else:
            if retry_count == 0:
                center_print(
                    text="Verification code is not valid. Please register again.",
                    text_type=DisplayBoxType.ERROR,
                )
                raise typer.Exit()

    # Register
    print("\n[yellow]Password must be at least 7 characters long[/yellow]")
    print(
        "[yellow]Password must contain at least one lowercase letter, one uppercase letter, and one special character (@, #, $, %, *, !)[/yellow]"
    )
    password = typer.prompt("> Enter password", hide_input=True)
    confirm_password = typer.prompt("> Confirm password", hide_input=True)
    if password != confirm_password:
        center_print(
            text="Password does not match. Please register again.",
            text_type=DisplayBoxType.ERROR,
        )
        raise typer.Exit()

    resp = auth_api.register(email=email, verification_code=verification_code, password=password)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    # Registration successful
    register_welcome()


@app.command(
    "login",
    help="[yellow]Login[/yellow] with email and password",
    rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]",
)
def login():
    email = typer.prompt("> Enter email address")
    password = typer.prompt("> Enter password", hide_input=True)

    resp = auth_api.login(email, password)
    if not resp.is_success:
        center_print(resp.error_message, DisplayBoxType.ERROR)
        raise typer.Exit()

    access_token = resp.body["accessToken"]
    AuthHandler.login(access_token)
    print("[yellow]Successfully logged in.[/yellow]")


@app.command(
    "logout",
    help="[red]Logout[/red] from the currently logged-in account",
    rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]",
)
def logout():
    AuthHandler.logout()
    print("[yellow]Successfully logged out.[/yellow]")


def register_welcome():
    print()
    with Progress() as progress:
        task = progress.add_task("[green]Registering...", total=100)

        while not progress.finished:
            progress.update(task, advance=1)
            time.sleep(0.02)

    welcome_msg = """[yellow] __      __          ___                                        
/\ \  __/\ \        /\_ \                                       
\ \ \/\ \ \ \     __\//\ \     ___    ___     ___ ___      __   
 \ \ \ \ \ \ \  /'__`\\\ \ \   /'___\ / __`\ /' __` __`\  /'__`\ 
  \ \ \_/ \_\ \/\  __/ \_\ \_/\ \__//\ \L\ \/\ \/\ \/\ \/\  __/ 
   \ `\___x___/\ \____\/\____\ \____\ \____/\ \_\ \_\ \_\ \____\\
    '\/__//__/  \/____/\/____/\/____/\/___/  \/_/\/_/\/_/\/____/
                                                                
                                                                
         __               ____                     __          __     
        /\ \__           /\  _`\   __             /\ \        /\ \    
        \ \ ,_\   ___    \ \ \L\_\/\_\  _ __   ___\ \ \/'\    \ \ \   
         \ \ \/  / __`\   \ \ \L_L\/\ \/\`'__\/ __`\ \ , <     \ \ \  
          \ \ \_/\ \L\ \   \ \ \/, \ \ \ \ \//\ \L\ \ \ \\\\\`\   \ \_\ 
           \ \__\ \____/    \ \____/\ \_\ \_\\\ \____/\ \_\ \_\   \/\_\\
            \/__/\/___/      \/___/  \/_/\/_/ \/___/  \/_/\/_/    \/_/
                                                              
                                                              [/yellow]"""
    l1 = " __      __          ___                                        "
    l2 = "/\ \  __/\ \        /\_ \                                       "
    l3 = "\ \ \/\ \ \ \     __\//\ \     ___    ___     ___ ___      __   "
    l4 = " \ \ \ \ \ \ \  /'__`\\\ \ \   /'___\ / __`\ /' __` __`\  /'__`\ "
    l5 = "  \ \ \_/ \_\ \/\  __/ \_\ \_/\ \__//\ \L\ \/\ \/\ \/\ \/\  __/ "
    l6 = "   \ `\___x___/\ \____\/\____\ \____\ \____/\ \_\ \_\ \_\ \____\\"
    l7 = "    '\/__//__/  \/____/\/____/\/____/\/___/  \/_/\/_/\/_/\/____/"

    r1 = "         __               ____                     __          __     "
    r2 = "        /\ \__           /\  _`\   __             /\ \        /\ \    "
    r3 = "        \ \ ,_\   ___    \ \ \L\_\/\_\  _ __   ___\ \ \/'\    \ \ \   "
    r4 = "         \ \ \/  / __`\   \ \ \L_L\/\ \/\`'__\/ __`\ \ , <     \ \ \  "
    r5 = "          \ \ \_/\ \L\ \   \ \ \/, \ \ \ \ \//\ \L\ \ \ \\\\\`\   \ \_\ "
    r6 = "           \ \__\ \____/    \ \____/\ \_\ \_\\\ \____/\ \_\ \_\   \/\_\\"
    r7 = "            \/__/\/___/      \/___/  \/_/\/_/ \/___/  \/_/\/_/    \/_/"

    #     console.print(Align.center("""[yellow] __      __          ___
    # /\ \  __/\ \        /\_ \
    # \ \ \/\ \ \ \     __\//\ \     ___    ___     ___ ___      __
    #  \ \ \ \ \ \ \  /'__`\\\ \ \   /'___\ / __`\ /' __` __`\  /'__`\
    #   \ \ \_/ \_\ \/\  __/ \_\ \_/\ \__//\ \L\ \/\ \/\ \/\ \/\  __/
    #    \ `\___x___/\ \____\/\____\ \____\ \____/\ \_\ \_\ \_\ \____\\
    #     '\/__//__/  \/____/\/____/\/____/\/___/  \/_/\/_/\/_/\/____/

    #          __               ____                     __          __
    #         /\ \__           /\  _`\   __             /\ \        /\ \
    #         \ \ ,_\   ___    \ \ \L\_\/\_\  _ __   ___\ \ \/'\    \ \ \
    #          \ \ \/  / __`\   \ \ \L_L\/\ \/\`'__\/ __`\ \ , <     \ \ \
    #           \ \ \_/\ \L\ \   \ \ \/, \ \ \ \ \//\ \L\ \ \ \\\\\`\   \ \_\
    #            \ \__\ \____/    \ \____/\ \_\ \_\\\ \____/\ \_\ \_\   \/\_\\
    #             \/__/\/___/      \/___/  \/_/\/_/ \/___/  \/_/\/_/    \/_/

    #                                                               [/yellow]"""))

    print("\n" * 5)
    screen_width = shutil.get_terminal_size().columns // 2
    with Live("", refresh_per_second=4) as live:
        num_iters = (70 * screen_width) // 109
        for _ in range(num_iters):
            space = _
            msg = (
                " " * space
                + l1
                + "\n"
                + " " * space
                + l2
                + "\n"
                + " " * space
                + l3
                + "\n"
                + " " * space
                + l4
                + "\n"
                + " " * space
                + l5
                + "\n"
                + " " * space
                + l6
                + "\n"
                + " " * space
                + l7
                + "\n"
            )
            live.update(Text(msg, style=Style(color="yellow")))
            time.sleep(0.009)
        print("\n")

    with Live("", refresh_per_second=4) as live:
        for _ in range(num_iters):
            space = _
            msg = (
                " " * space
                + r1
                + "\n"
                + " " * space
                + r2
                + "\n"
                + " " * space
                + r3
                + "\n"
                + " " * space
                + r4
                + "\n"
                + " " * space
                + r5
                + "\n"
                + " " * space
                + r6
                + "\n"
                + " " * space
                + r7
                + "\n"
            )
            live.update(Text(msg, style=Style(color="yellow")))
            time.sleep(0.009)

    print("\n" * 1)
    console.print(
        " " * ((70 * screen_width) // 109 + 20) + "Enter [green]girok login[/green] to log-in to your account."
    )
    print("\n" * 10)



from collections import defaultdict
from typing import List, Dict, Optional
import asyncio
import calendar
from calendar import monthrange
from datetime import datetime, timedelta

from rich.markdown import Markdown
from rich.panel import Panel
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import log
from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.messages import Message
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Placeholder,
    Static,
    Tree,
)

import girok.api.category as category_api
import girok.api.task as task_api
import girok.calendar_cli.utils as calendar_utils
from girok.calendar_cli.sidebar import CategoryTree
from girok.utils.time import get_year_and_month_by_month_offset
from girok.constants import CALENDAR_HEADER_DATE_COLOR, CALENDAR_TODAY_COLOR, CALENDAR_WEEKDAY_NAME_COLOR
from girok.calendar_cli.entity import Category
from girok.commands.task.command import map_to_event_entities
from girok.commands.task.entity import Event
from girok.utils.time import convert_iso_date_str_to_date_obj


class WeekdayBarContainer(Horizontal):
    pass


class CalendarHeader(Vertical):
    year = datetime.now().year
    month = datetime.now().month
    cat_path = ""
    category: Category = None
    tag = ""

    def on_mount(self):
        self.display_date()

    def compose(self):
        month_name = calendar.month_name[self.month]

        with Horizontal():
            with Container(id="calendar-header-category-container"):
                yield Static(self.cat_path, id="calendar-header-category")
            with Container(id="calendar-header-date-container"):
                yield Static(
                    Text(
                        f"{month_name} {self.year}",
                        style=Style(
                            color=CALENDAR_HEADER_DATE_COLOR, bold=True
                        ),
                    ),
                    id="calendar-header-date",
                )
            with Container(id="calendar-header-tag-container"):
                yield Static(f"{self.tag}", id="calendar-header-tag")
        yield Horizontal()
        with WeekdayBarContainer(id="weekday-bar"):
            yield Static(
                Text(
                    "Monday",
                    style=Style(color=CALENDAR_WEEKDAY_NAME_COLOR, bold=True),
                ),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text(
                    "Tuesday",
                    style=Style(color=CALENDAR_WEEKDAY_NAME_COLOR, bold=True),
                ),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text(
                    "Wednesday",
                    style=Style(color=CALENDAR_WEEKDAY_NAME_COLOR, bold=True),
                ),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text(
                    "Thursday",
                    style=Style(color=CALENDAR_WEEKDAY_NAME_COLOR, bold=True),
                ),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text(
                    "Friday",
                    style=Style(color=CALENDAR_WEEKDAY_NAME_COLOR, bold=True),
                ),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text("Saturday", style=Style(color="#87C5FA", bold=True)),
                classes="calendar-weekday-name",
            )
            yield Static(
                Text("Sunday", style=Style(color="#DB4455", bold=True)),
                classes="calendar-weekday-name",
            )

    def update_year_and_month(self, year, month):
        self.year, self.month = year, month
        self.display_date()

    def update_category(self, category: Category):
        self.cat_path = category.path
        self.display_date()

    def update_tag(self, new_tag: Optional[str]):
        self.tag = new_tag if new_tag else "All"
        self.display_date()

    def display_date(self):
        month_name = calendar.month_name[self.month]
        calendar_header_category = self.query_one("#calendar-header-category")
        calendar_header_date = self.query_one("#calendar-header-date")
        calendar_header_tag = self.query_one("#calendar-header-tag")
        calendar_weekday_bar = self.query_one("#weekday-bar")

        calendar_header_category.update(
            Text(
                f"Category: /{self.cat_path}",
                style=Style(color=CALENDAR_HEADER_DATE_COLOR),
            )
        )
        calendar_header_date.update(
            Text(
                f"{month_name} {self.year}",
                style=Style(color=CALENDAR_HEADER_DATE_COLOR, bold=True),
            )
        )
        calendar_header_tag.update(
            Text(
                f"Tag: {self.tag}",
                style=Style(color=CALENDAR_HEADER_DATE_COLOR),
            )
        )


class CalendarCell(Vertical):
    pass


class Calendar(Container):
    year = datetime.now().year
    month = datetime.now().month
    cat_path = ""  # If "", show all categories
    cat_id: int = None
    category: Category = None
    tag = ""
    tasks = []
    events: List[Event] = []
    can_focus = True
    cur_month_first_day_cell_num = None
    cur_focused_cell_cord = (None, None)
    cur_focused_cell = None
    day_to_events_map: Dict[str, List[Event]] = defaultdict(list)
    m = 5
    n = 7
    grid = [[False for _ in range(7)] for _ in range(5)]
    is_pop_up = False

    class TaskCellSelected(Message):
        def __init__(self, cell_events: List[Event], year: int, month: int, day: int):
            super().__init__()
            self.cell_events = cell_events
            self.year = year
            self.month = month
            self.day = day

    def on_mount(self):
        self.update_calendar()

    def compose(self):
        for i in range(35):
            yield CalendarCell(classes="calendar-cell", id=f"cell{i}")

    def on_key(self, event):
        if self.is_pop_up:
            return
        x, y = self.cur_focused_cell_cord
        if event.key == "h":  # left
            next_cell_coord = (x, y - 1)
        elif event.key == "j":  # down
            next_cell_coord = (x + 1, y)
        elif event.key == "k":  # up
            next_cell_coord = (x - 1, y)
        elif event.key == "l":  # right
            next_cell_coord = (x, y + 1)
        elif event.key == "o":
            pass
        else:
            return

        if event.key in ["h", "j", "k", "l"]:  # moving on cells
            nx, ny = next_cell_coord
            if nx < 0 or ny < 0 or nx >= self.m or ny >= self.n:  # Out of matrix
                return

            if not self.grid[nx][ny]:  # Out of boundary of the current month
                return

            prev_cell_num = calendar_utils.convert_coord_to_cell_num(
                *self.cur_focused_cell_cord
            )
            prev_cell = self.query_one(f"#cell{prev_cell_num}")
            calendar_utils.remove_left_arrow(prev_cell)

            cur_cell_num = calendar_utils.convert_coord_to_cell_num(nx, ny)
            next_cell = self.query_one(f"#cell{cur_cell_num}")
            calendar_utils.add_left_arrow(next_cell)

            self.cur_focused_cell_cord = (nx, ny)
            self.cur_focused_cell = next_cell
        elif event.key == "o":  # select a cell
            cur_cell_num = calendar_utils.convert_coord_to_cell_num(x, y)
            cur_cell = self.query_one(f"#cell{cur_cell_num}")
            # cell_events = cur_cell.children[1:]  # task data

            # Retrieve tasks for the selected day
            selected_day = calendar_utils.convert_cell_num_to_day(
                self.year, self.month, cur_cell_num
            )

            self.post_message(
                self.TaskCellSelected(self.day_to_events_map[selected_day], self.year, self.month, selected_day)
            )
            self.is_pop_up = True

    def on_focus(self):
        x, y = calendar_utils.convert_cell_num_to_coord(
            self.cur_month_first_day_cell_num
        )
        target_cell = self.query_one(f"#cell{self.cur_month_first_day_cell_num}")
        calendar_utils.add_left_arrow(target_cell)
        self.cur_focused_cell_cord = (x, y)
        self.cur_focused_cell = target_cell

    def update_year_and_month(self, year, month):
        self.year, self.month = year, month
        self.update_calendar()

    def update_category(self, category: Category):
        self.cat_path = category.path
        self.cat_id = category.id
        self.category = category
        self.update_calendar(show_arrow=False)

    def update_tag(self, new_tag: Optional[str]):
        self.tag = new_tag
        self.update_calendar(show_arrow=False)

    def refresh_cell_days(self):
        self.grid = [[False for _ in range(7)] for _ in range(5)]
        first_weekday, total_days = calendar.monthrange(self.year, self.month)
        self.cur_month_first_day_cell_num = first_weekday
        now = datetime.now()
        for i in range(35):
            cell = self.query_one(f"#cell{i}")
            for child in cell.walk_children():
                child.remove()
            if i >= first_weekday and i <= first_weekday + total_days - 1:
                x, y = calendar_utils.convert_cell_num_to_coord(i)
                self.grid[x][y] = True
                day = calendar_utils.convert_cell_num_to_day(self.year, self.month, i)
                day_text = Text()
                day_text.append(f"{day}")
                if self.year == now.year and self.month == now.month and day == now.day:
                    day_text = Text(
                        str(day_text),
                        style=Style(
                            bgcolor=CALENDAR_TODAY_COLOR, color="black"
                        ),
                    )
                cell.mount(Label(day_text, id=f"cell-header-{i}"))

    def update_calendar(self, show_arrow=True):
        """
        If val == "", then "root category" is selected
        """

        # Filter by category
        category_id = None
        if self.category:
            category_id = self.category.id

        # Filter by tag
        tags = None if not self.tag else [self.tag]

        # Filter by date
        current_month_first_date = datetime(self.year, self.month, 1)
        current_month_last_day = monthrange(self.year, self.month)[1]
        current_month_last_date = datetime(self.year, self.month, current_month_last_day)

        start_date = current_month_first_date.strftime("%Y-%m-%d")
        end_date = current_month_last_date.strftime("%Y-%m-%d")

        # Retrieve all events
        resp = task_api.get_all_tasks(start_date=start_date, end_date=end_date, category_id=category_id, tags=tags, fetch_children=True)
        if not resp.is_success:
            exit(0)
        self.events = map_to_event_entities(resp.body['events'])

        # Empty out current calendar view
        self.refresh_cell_days()

        first_weekday, last_day = calendar.monthrange(self.year, self.month)

        if self.cur_focused_cell:
            if show_arrow:
                calendar_utils.remove_left_arrow(self.cur_focused_cell)
            self.cur_focused_cell_cord = calendar_utils.convert_cell_num_to_coord(
                first_weekday
            )
            self.cur_focused_cell = self.query_one(f"#cell{first_weekday}")  # update
            if show_arrow:
                calendar_utils.add_left_arrow(self.cur_focused_cell)

        for idx, event in enumerate(self.events):
            target_days = event.get_all_days_for_month(self.year, self.month)
            for day in target_days:
                self.day_to_events_map[day].append(event)
                # event_date_obj = convert_iso_date_str_to_date_obj(event.event_date.start_date)
                # day = event_date_obj.day
                cell_num = calendar_utils.convert_day_to_cell_num(
                    self.year, self.month, day
                )
                cell = self.query_one(f"#cell{cell_num}")
                color = event.color_hex
                name = event.name
                # if len(name) > 13:
                #     name = name[:13] + ".."
                task_item_name = Text()
                task_item_name.append("●", style=color)
                task_item_name.append(" " + name)

                task_item = Static(
                    task_item_name, id=f"task-cell{cell_num}-{idx}", classes="task-item"
                )
                cell.mount(task_item)

                task_item = self.query_one(f"#task-cell{cell_num}-{idx}")
                task_item.styles.overflow_x = "hidden"
                task_item.styles.overflow_y = "hidden"


class CalendarContainer(Vertical):
    year = datetime.now().year
    month = datetime.now().month
    tag = None

    def update_month_by_offset(self, offset: int):
        new_year, new_month = get_year_and_month_by_month_offset(
            month_offset=offset, year=self.year, month=self.month
        )
        self.year, self.month = new_year, new_month
        calendar_header = self.query_one(CalendarHeader)
        cal = self.query_one(Calendar)

        calendar_header.update_year_and_month(self.year, self.month)
        cal.update_year_and_month(self.year, self.month)

    def update_year_and_month(self, year: int, month: int):
        self.year, self.month = year, month
        calendar_header = self.query_one(CalendarHeader)
        cal = self.query_one(Calendar)

        calendar_header.update_year_and_month(self.year, self.month)
        cal.update_year_and_month(self.year, self.month)

    def update_category(self, category: Category):
        """
        cat_path: ex) HKU/COMP3230 or ""
        """
        cal = self.query_one(Calendar)
        cal.update_category(category)

        cal_header = self.query_one(CalendarHeader)
        cal_header.update_category(category)

    def update_tag(self, tag: Optional[str]):
        self.tag = tag
        cal = self.query_one(Calendar)
        cal.update_tag(new_tag=tag)
        cal_header = self.query_one(CalendarHeader)
        cal_header.update_tag(new_tag=tag)

    def compose(self):
        yield CalendarHeader(id="calendar-header")
        yield Calendar(id="calendar")
